import os
import unittest

from zope.testing import doctest
from zope import component
from zope import interface
from zope.component import testing
from zope.app.testing.functional import ZCMLLayer
import zope.traversing.adapters
import zope.traversing.namespace
import zope.publisher.interfaces.browser
import z3c.form.testing

import plone.z3cform.templates

def create_eventlog(event=interface.Interface):
    value = []
    @component.adapter(event)
    def log(event):
        value.append(event)
    component.provideHandler(log)
    return value

def setup_defaults():
    # Set up z3c.form defaults
    z3c.form.testing.setupFormDefaults()
    # Make traversal work; register both the default traversable
    # adapter and the ++view++ namespace adapter
    component.provideAdapter(
        zope.traversing.adapters.DefaultTraversable, [None])
    component.provideAdapter(
        zope.traversing.namespace.view, (None, None), name='view')

    # Setup ploneform macros
    component.provideAdapter(
        plone.z3cform.templates.Macros,
        (None, None),
        zope.publisher.interfaces.browser.IBrowserView,
        name='ploneform-macros')
    # setup plone.z3cform template
    from zope.pagetemplate.interfaces import IPageTemplate

    component.provideAdapter(
        plone.z3cform.templates.form_factory,
        (None, None),
        IPageTemplate)

    component.provideAdapter(
        z3c.form.error.ErrorViewSnippet,
        (None, None, None, None, None, None),
        z3c.form.interfaces.IErrorViewSnippet)

testing_zcml_path = os.path.join(os.path.dirname(__file__), 'testing.zcml')
testing_zcml_layer = ZCMLLayer(
    testing_zcml_path, 'plone.z3cform', 'testing_zcml_layer')

def test_suite():
    layout_txt = doctest.DocFileSuite('layout.txt')
    layout_txt.layer = testing_zcml_layer
    
    fieldsets_txt = doctest.DocFileSuite('fieldsets/README.txt')
    fieldsets_txt.layer = testing_zcml_layer
    
    return unittest.TestSuite([
        layout_txt, fieldsets_txt,

        doctest.DocFileSuite(
           'README.txt',
           setUp=testing.setUp, tearDown=testing.tearDown,
           ),

        doctest.DocFileSuite(
           'crud/README.txt',
           setUp=testing.setUp, tearDown=testing.tearDown,
           ),

        doctest.DocTestSuite(
           'plone.z3cform.crud.crud',
           setUp=testing.setUp, tearDown=testing.tearDown,
           ),
        ])



from ZODB.POSException import ConflictError
from zope import interface
import zope.event
import zope.lifecycleevent
from z3c.form import button
from z3c.form import field
from z3c.form import form
import z3c.form.widget
from z3c.form.interfaces import DISPLAY_MODE, INPUT_MODE, NOVALUE
from zope.app.pagetemplate import viewpagetemplatefile

from plone.z3cform.widget import singlecheckboxwidget_factory
from plone.z3cform import MessageFactory as _


class ICrudForm(interface.Interface):

    update_schema = interface.Attribute(
        "Editable part of the schema for use in the update form.")

    view_schema = interface.Attribute(
        "Viewable (only) part of the schema for use in the update form.")

    add_schema = interface.Attribute(
        "Schema for use in the add form; defaults to ``update_schema``.")

    def get_items():
        """Subclasses must a list of all items to edit.

        This list contains tuples of the form ``(id, item)``, where
        the id is a unique identifiers to the items.  The items must
        be adaptable to the schema returned by ``update_schema`` and
        ``view_schema`` methods.
        """

    def add(data):
        """Subclasses must implement this method to create an item for
        the given `data` *and* add it to a container, and return it.

        The `data` mapping corresponds to the schema returned by
        `add_schema`.
        """

    def remove((id, item)):
        """Subclasses must implement this method to remove the given
        item from the site.

        It's left to the implementing class to notify of
        ``zope.app.container.contained.ObjectRemovedEvent``.
        """

    def before_update(item, data):
        """A hook that gets called before an item is updated.
        """

    def link(item, field):
        """Return a URL for this item's field or None.
        """

class AbstractCrudForm(object):
    """The AbstractCrudForm is not a form but implements methods
    necessary to comply with the ``ICrudForm`` interface:
    
      >>> from zope.interface.verify import verifyClass
      >>> verifyClass(ICrudForm, AbstractCrudForm)
      True
    """
    interface.implements(ICrudForm)

    update_schema = None
    view_schema = None

    @property
    def add_schema(self):
        return self.update_schema

    def get_items(self):
        raise NotImplementedError

    def add(self, data):
        raise NotImplementedError

    def remove(self, (id, item)):
        raise NotImplementedError

    def before_update(self, item, data):
        pass

    def link(self, item, field):
        return None

class EditSubForm(form.EditForm):
    template = viewpagetemplatefile.ViewPageTemplateFile('crud-row.pt')

    @property
    def prefix(self):
        return 'crud-edit.%s.' % self.content_id

    # These are set by the parent form
    content = None
    content_id = None

    @property
    def fields(self):
        fields = field.Fields(self._select_field())

        crud_form = self.context.context

        update_schema = crud_form.update_schema
        if update_schema is not None:
            fields += field.Fields(update_schema)

        view_schema = crud_form.view_schema
        if view_schema is not None:
            view_fields = field.Fields(view_schema)
            for f in view_fields.values():
                f.mode = DISPLAY_MODE
                # This is to allow a field to appear in both view
                # and edit mode at the same time:
                if not f.__name__.startswith('view_'):
                    f.__name__ = 'view_' + f.__name__
            fields += view_fields
            
        return fields

    def getContent(self):
        return self.content

    def _select_field(self):
        select_field = field.Field(
            zope.schema.Bool(__name__='select',
                             required=False,
                             title=_(u'select')))
        select_field.widgetFactory[INPUT_MODE] = singlecheckboxwidget_factory
        return select_field
        
    # XXX: The three following methods, 'getCombinedWidgets',
    # 'getTitleWidgets', and 'getNiceTitles' are hacks to support the
    # page template.  Let's get rid of them.
    def getCombinedWidgets(self):
        """Returns pairs of widgets to improve layout"""
        widgets = self.widgets.items()
        combined = []
        seen = set()
        for name, widget in list(widgets):
            if widget.mode == INPUT_MODE:
                view_widget = self.widgets.get('view_%s' % name)
                if view_widget is not None:
                    combined.append((widget, view_widget))
                    seen.add(view_widget)
                else:
                    combined.append((widget,))
            else:
                if widget not in seen:
                    combined.append((widget,))
        return combined
             
    def getTitleWidgets(self):
        combinedWidgets = self.getCombinedWidgets()
        widgetsForTitles = [w[0] for w in combinedWidgets]
        return widgetsForTitles

    def getNiceTitles(self):
        widgetsForTitles = self.getTitleWidgets()        
        freakList = []
        for item in widgetsForTitles:
            freakList.append(item.field.title)
        return freakList
        
class EditForm(form.Form):
    label = _(u"Edit")
    template = viewpagetemplatefile.ViewPageTemplateFile('crud-table.pt')
        
    @property
    def prefix(self):
        parent_prefix = getattr(self.context, 'prefix', '')
        return 'crud-edit.' + parent_prefix

    def update(self):
        self._update_subforms()
        super(EditForm, self).update()
    
    def _update_subforms(self):
        self.subforms = []
        for id, item in self.context.get_items():
            subform = EditSubForm(self, self.request)
            subform.content = item
            subform.content_id = id
            subform.update()
            self.subforms.append(subform)

    @button.buttonAndHandler(_('Apply changes'), name='edit')
    def handle_edit(self, action):
        success = _(u"Successfully updated")
        partly_success = _(u"Some of your changes could not be applied.")
        status = no_changes = _(u"No changes made.")
        for subform in self.subforms:
            # With the ``extractData()`` call, validation will occur,
            # and errors will be stored on the widgets amongst other
            # places.  After this we have to be extra careful not to
            # call (as in ``__call__``) the subform again, since
            # that'll remove the errors again.  With the results that
            # no changes are applied but also no validation error is
            # shown.
            data, errors = subform.extractData()
            if errors:
                if status is no_changes:
                    status = subform.formErrorsMessage
                elif status is success:
                    status = partly_success
                continue
            del data['select']
            self.context.before_update(subform.content, data)
            changes = subform.applyChanges(data)
            if changes:
                if status is no_changes:
                    status = success
                elif status is subform.formErrorsMessage:
                    status = partly_success

                # If there were changes, we'll update the view widgets
                # again, so that they'll actually display the changes
                [w.update() for w in subform.widgets.values()
                 if w.mode == DISPLAY_MODE]
        self.status = status

    @button.buttonAndHandler(_('Delete'), name='delete')
    def handle_delete(self, action):
        selected = self.selected_items()
        if selected:
            self.status = _(u"Successfully deleted items.")
            for id, item in selected:
                try:
                    self.context.remove((id, item))
                except ConflictError:
                    raise
                except:
                    # In case an exception is raised, we'll catch it
                    # and notify the user; in general, this is
                    # unexpected behavior:
                    self.status = _(u'Unable to remove one or more items.')
                    break
        
            # We changed the amount of entries, so we update the subforms again.
            self._update_subforms()
        else:
            self.status = _(u"Please select items to delete.")

    def selected_items(self):
        tuples = []
        for subform in self.subforms:
            data = subform.widgets['select'].extract()
            if not data or data is NOVALUE:
                continue
            else:
                tuples.append((subform.content_id, subform.content))
        return tuples

class AddForm(form.Form):
    label = _(u"Add")
    template = viewpagetemplatefile.ViewPageTemplateFile('../form.pt')
    ignoreContext = True
    ignoreRequest = True

    @property
    def prefix(self):
        parent_prefix = getattr(self.context, 'prefix', '')
        return 'crud-add.' + parent_prefix

    @property
    def fields(self):
        return field.Fields(self.context.add_schema)

    @button.buttonAndHandler(_('Add'), name='add')
    def handle_add(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = form.AddForm.formErrorsMessage
            return
        item = self.context.add(data)
        zope.event.notify(zope.lifecycleevent.ObjectCreatedEvent(item))
        self.status = _(u"Item added successfully.")

class NullForm(object):
    def __init__(self, context, request):
        self.context = context
        self.request = request

    def update(self):
        pass

    def render(self):
        return ''
    __call__ = render

class CrudForm(AbstractCrudForm, form.Form):
    template = viewpagetemplatefile.ViewPageTemplateFile('crud-master.pt')
    description = u''

    editform_factory = EditForm
    addform_factory = AddForm

    def update(self):
        super(CrudForm, self).update()

        addform = self.addform_factory(self, self.request)
        editform = self.editform_factory(self, self.request)
        addform.update()
        editform.update()
        self.subforms = [editform, addform]



from z3c.form.field import Fields
from z3c.form.util import expandPrefix

from plone.z3cform.fieldsets.group import GroupFactory

def add(form, *args, **kwargs):
    """Add one or more fields. Keyword argument 'index' can be used to
    specify an index to insert at. Keyword argument 'group' can be used
    to specify the label of a group, which will be found and used or
    created if it doesn't exist.
    """
    
    index = kwargs.pop('index', None)
    group = kwargs.pop('group', None)
    
    new_fields = Fields(*args, **kwargs)

    if not group or isinstance(group, basestring):
        source = find_source(form, group=group)
    else:
        source = group

    if source is None and group:
        source = GroupFactory(group, new_fields)
        form.groups.append(source)
    else:        
        if index is None or index >= len(source.fields):
            source.fields += new_fields
        else:
            field_names = source.fields.keys()
            source.fields = source.fields.select(*field_names[:index]) + \
                            new_fields + \
                            source.fields.select(*field_names[index:])

def remove(form, field_name, prefix=None):
    """Get rid of a field. The omitted field will be returned.
    """
    
    if prefix:
        field_name = expandPrefix(prefix) + field_name
    
    if field_name in form.fields:
        field = form.fields[field_name]
        form.fields = form.fields.omit(field_name)
        return field
    else:
        for group in form.groups:
            if field_name in group.fields:
                field = group.fields[field_name]
                group.fields = group.fields.omit(field_name)
                return field

def move(form, field_name, before=None, after=None, prefix=None, relative_prefix=None):
    """Move the field with the given name before or after another field.
    """
    
    if prefix:
        field_name = expandPrefix(prefix) + field_name
    
    if before and after:
        raise ValueError(u"Only one of 'before' or 'after' is allowed")
    
    offset = 0
    if after:
        offset = 1
    
    relative = before or after
    if relative_prefix:
        relative = expandPrefix(relative_prefix) + relative
    
    if field_name not in form.fields:
        raise KeyError("Field %s not found" % field_name)
    
    if relative not in form.fields:
        found = False
        for group in form.groups:
            if relative in group.fields:
                found = True
                break
        if not found:
            raise KeyError("Field %s not found" % relative)
    
    field = remove(form, field_name)
    
    group = None
    index = None
    
    if relative in form.fields:
        index = form.fields.keys().index(relative)
    else:
        for group in form.groups:
            if relative in group.fields:
                index = group.fields.keys().index(relative)
                break
    
    if index is None:
        raise KeyError("Field %s not found" % relative)
    
    add(form, field, group=group, index=index+offset)

def find_source(form, group=None):
    if group is not None:
        group_factory = [g for g in form.groups if g.__name__ == group]
        if len(group_factory) >= 1:
            return group_factory[0]
        else:
            return None
    return form


from zope.interface import Interface
from zope import schema

from z3c.form.interfaces import IFields, IGroup, IGroupForm

class IFormExtender(Interface):
    """A component that can add, modify, sort and group fields in a form.
    
    This should be a named multi adapter from (context, request, form).
    """
    
    order = schema.Int(title=u"Order",
                       description=u"Use this property to order the sorter. " +
                                   u"Low numbers are executed before high ones.",
                       required=True)
    
    def update():
        """Modify the form in place. Supported operations include:
        
         - modify the 'fields' object to change the default fieldset
         - modify the 'groups' list to add, remove or reorder fieldsets
         - modify the 'fields' property of a given group
        """

class IDescriptiveGroup(IGroup):
    """Extension to z3c.form's Group class that can separate out a name,
    a label and a description.
    """
    
    __name__ = schema.TextLine(title=u"Name of this group")
    
    label = schema.TextLine(title=u"Fieldset title",
                            description=u"The __name__ will be used if this is not given",
                            required=False)
                            
    description = schema.Text(title=u"Fieldset description",
                              required=False)

class IGroupFactory(Interface):
    """An object that can be used to create a z3c.form.group.Group.
    """
    
    __name__ = schema.TextLine(title=u"Name of this group")
    
    label = schema.TextLine(title=u"Fieldset title",
                            description=u"The __name__ will be used if this is not given",
                            required=False)
                            
    description = schema.Text(title=u"Fieldset description",
                              required=False)
    
    fields = schema.Object(title=u"Fields in this form", schema=IFields)

class IExtensibleForm(Interface):
    """A special version of the IGroupForm that is extensible via
    IFormExtender adapters.
    """

    groups = schema.List(title=u'Groups',
                         value_type=schema.Object(title=u"Group", schema=IGroupFactory))

    default_fieldset_label = schema.TextLine(title=u"Label of the default fieldset")
    
    def updateFields():
        """Called during form update to allow updating of self.fields
        and self.groups.
        """


from zope import interface
from zope.i18n.interfaces import IUserPreferredLanguages
from zope.i18n.locales import locales, LoadLocaleError
from Products.Five.browser import decode
import z3c.form.interfaces

class IFixedUpRequest(interface.Interface):
    pass

# XXX This is ripped from zope.publisher.http.HTTPRequest; we should
# move this into Five
def setup_locale(request):
    envadapter = IUserPreferredLanguages(request, None)
    if envadapter is None:
        return None
    
    langs = envadapter.getPreferredLanguages()
    for httplang in langs:
        parts = (httplang.split('-') + [None, None])[:3]
        try:
            return locales.getLocale(*parts)
        except LoadLocaleError:
            # Just try the next combination
            pass
    else:
        # No combination gave us an existing locale, so use the default,
        # which is guaranteed to exist
        return locales.getLocale(None, None, None)

# XXX Add a getURL method on the request object; we should move this
# into Five
def add_getURL(request):
    def getURL(level=0, path_only=False):
        assert level == 0 and path_only == False
        return request['ACTUAL_URL']
    request.getURL = getURL

def switch_on(view, request_layer=z3c.form.interfaces.IFormLayer):
    request = view.request
    from zope.publisher.interfaces.browser import IBrowserApplicationRequest
    if not IFixedUpRequest.providedBy(request) and \
       not IBrowserApplicationRequest.providedBy(request):
        interface.alsoProvides(request, IFixedUpRequest)
        interface.alsoProvides(request, request_layer)
        request.locale = setup_locale(request)
        add_getURL(request)
        decode.processInputs(request)



from zope.interface import implements

from plone.z3cform.fieldsets.interfaces import IGroupFactory
from plone.z3cform.fieldsets.interfaces import IDescriptiveGroup

from z3c.form import group

class Group(group.Group):
    implements(IDescriptiveGroup)
    
    __name__ = u""
    label = u""
    description = u""

class GroupFactory(object):
    implements(IGroupFactory)
    
    def __init__(self, __name__, fields, label=None, description=None):
        self.__name__ = __name__
        self.fields = fields
        
        self.label = label or __name__
        self.description = description
        
    def __call__(self, context, request, parentForm):
        g = Group(context, request, parentForm)
        g.__name__ = self.__name__
        g.label = self.label
        g.description = self.description
        g.fields = self.fields
        return g


from zope import interface
from zope.schema import vocabulary

import z3c.form.term
import z3c.form.browser.checkbox
import z3c.form.interfaces

class SingleCheckboxWidget(z3c.form.browser.checkbox.SingleCheckBoxWidget):
    """XXX: We need to refactor this and patch z3c.form where
    it makes sense.
    """

    def update(self):
        self.ignoreContext = True
        super(SingleCheckboxWidget, self).update()

    def updateTerms(self):
        # The default implementation would render "selected" as a
        # lebel for the single checkbox.  We use no label instead.
        if self.terms is None:
            self.terms = z3c.form.term.Terms()
            self.terms.terms = vocabulary.SimpleVocabulary((
                vocabulary.SimpleTerm(True, 'selected', u''),
                ))
        return self.terms

    def extract(self, default=z3c.form.interfaces.NOVALUE):
        # The default implementation returns [] here.
        if (self.name not in self.request and
            self.name+'-empty-marker' in self.request):
            return default
        else:
            return super(SingleCheckboxWidget, self).extract(default)

@interface.implementer(z3c.form.interfaces.IDataConverter)
def singlecheckboxwidget_factory(field, request):
    widget = SingleCheckboxWidget(request)
    return z3c.form.widget.FieldWidget(field, widget)



from setuptools import setup, find_packages
import os

version = '0.3'


def description():
    join = lambda *paths: os.path.join('plone', 'z3cform', *paths)
    return (open('README.txt').read() + '\n' +
            open(join('README.txt')).read() + '\n' +
            open(join('crud', 'README.txt')).read() + '\n' +
            open(join('fieldsets', 'README.txt')).read() + '\n' +
            open(os.path.join('docs', 'HISTORY.txt')).read() + '\n')

setup(name='plone.z3cform',
      version=version,
      description="A library that allows use of z3c.form with Zope 2 "
      "(and Plone)",
      long_description=description(),
      classifiers=[
        "Framework :: Plone",
        "Framework :: Zope2",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        ],
      keywords='zope cmf form widget',
      author='Daniel Nouri and contributors',
      author_email='daniel.nouri@gmail.com',
      url='http://pypi.python.org/pypi/plone.z3cform',
      license='ZPL 2.1',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['plone'],
      include_package_data=True,
      zip_safe=False,

      # If in Zope 2, z3c.form or another Zope 3 package starts
      # pulling incompatible dependencies, use the "fake zope eggs"
      # feature of plone.recipe.zope2install.
      install_requires=[
          'setuptools',
          'z3c.form',
          'zope.i18n>=3.4'
      ],
      )



"""This module provides "form template factories" that can be
registered to provide default form templates for forms and subforms
that have a Plone style.  These default templates draw from a macro
page template which you can use by itself to render parts of it.
"""

import os
import zope.publisher.browser
import zope.app.pagetemplate.viewpagetemplatefile

import z3c.form.interfaces
from z3c.form.form import FormTemplateFactory

import plone.z3cform

path = lambda p: os.path.join(os.path.dirname(plone.z3cform.__file__), p)

form_factory = FormTemplateFactory(
    path('form.pt'), form=z3c.form.interfaces.IForm)
subform_factory = FormTemplateFactory(
    path('subform.pt'), form=z3c.form.interfaces.ISubForm)

class Macros(zope.publisher.browser.BrowserView):
    template = zope.app.pagetemplate.viewpagetemplatefile.ViewPageTemplateFile(
        'macros.pt')
    
    def __getitem__(self, key):
        return self.template.macros[key]



import z3c.form.interfaces
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile

from plone.z3cform import z2

class FormWrapper(BrowserView):
    """Use this as a base class for your Five view and override the
    'form' attribute with your z3c.form form class.  Your form will
    then be rendered in the contents area of a layout template, the
    'index' attribute.

    Use the 'wrap' function in this module if you don't like defining
    classes.
    """
    index = ViewPageTemplateFile('layout.pt')
    form = None # override this with a form class.
    request_layer = z3c.form.interfaces.IFormLayer
    
    def __call__(self):
        """This method renders the outer skeleton template, which in
        turn calls the 'contents' method below.

        We use an indirection to 'self.index' here to allow users to
        override the skeleton template through the 'browser' zcml
        directive.
        """
        return self.index()

    def contents(self):
        """This is the method that'll call your form.  You don't
        usually override this.
        """
        # A call to 'switch_on' is required before we can render
        # z3c.forms within Zope 2.
        z2.switch_on(self, request_layer=self.request_layer)
        return self.render_form()

    def render_form(self):
        """This method returns the rendered z3c.form form.

        Override this method if you need to pass a different context
        to your form, or if you need to render a number of forms.
        """
        return self.form(self.context.aq_inner, self.request)()

    def label(self):
        """Override this method to use a different way of acquiring a
        label or title for your page.  Overriding this with a simple
        attribute works as well.
        """
        return self.form.label

def wrap_form(form, __wrapper_class=FormWrapper, **kwargs):
    class MyFormWrapper(__wrapper_class):
        pass
    MyFormWrapper.form = form
    for name, value in kwargs.items():
        setattr(MyFormWrapper, name, value)
    return MyFormWrapper



from zope.interface import implements
from zope.component import getAdapters

from plone.z3cform.fieldsets.interfaces import IExtensibleForm
from plone.z3cform.fieldsets.interfaces import IFormExtender

from z3c.form.group import GroupForm

from plone.z3cform.fieldsets import utils

from plone.z3cform import MessageFactory as _

def order_key(adapter_tuple):
    return adapter_tuple[1].order

class FormExtender(object):
    """Base class for IFormExtender adapters with convenience APIs
    """
    implements(IFormExtender)
    
    # Change this to prioritise
    order = 0
    
    def __init__(self, context, request, form):
        self.context = context
        self.request = request
        self.form = form
    
    def add(self, *args, **kwargs):
        """Add one or more fields. Keyword argument 'index' can be used to
        specify an index to insert at. Keyword argument 'group' can be used
        to specify the label of a group, which will be found and used or
        created if it doesn't exist.
        """
        
        utils.add(self.form, *args, **kwargs)
    
    def remove(self, field_name, prefix=None):
        """Get rid of a field. The omitted field will be returned.
        """
        
        return utils.remove(self.form, field_name, prefix=prefix)
    
    def move(self, field_name, before=None, after=None, prefix=None, relative_prefix=None):
        """Move the field with the given name before or after another field.
        """
        
        utils.move(self.form, field_name, before=before, after=after,
                    prefix=prefix, relative_prefix=relative_prefix)

class ExtensibleForm(GroupForm):
    implements(IExtensibleForm)
    
    groups = []
    default_fieldset_label = _(u"Default")
    
    def update(self):
        self.updateFields()
        super(ExtensibleForm, self).update()
        
    def updateFields(self):
        extenders = getAdapters((self.context, self.request, self), IFormExtender)
        for name, extender in sorted(extenders, key=order_key):
            extender.update()